#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "music.h"
#include "m_irq.h"
#include "adldata.h" /* for Period[] */

/* 
 Supported commandinfobytes currently include:
 
  Axx (set speed (ticks))
  Bxx
  Cxx
  Dx0 (volume slide up   ticks*x)
  D0y (volume slide down ticks*y)
  DxF (increase volume by y)
  DFy (decrease volume by y)
  DFF (maximizes volume)

  EFx (freqdec)           ("jotain sinne pin", difficult to emulate)
  Exx (freqslide)         ("jotain sinne pin", difficult to emulate)

  FFx (freqinc)           ("jotain sinne pin", difficult to emulate)
  Fxx (freqslide)         ("jotain sinne pin", difficult to emulate)

  Hxx (incorrect yet, difficult to emulate)
  Kxx (incorrect yet, difficult to emulate)

  Jxx (arpeggio)
  Pxx (NON-STANDARD! Specify arpeggio speed - default=32h (50Hz))
      (If bit 7 set, then arpeggio is only with two notes, not three.)

  Xxx (panning, not fully supported for FM. 00..FF, 80h=middle)

  Qxy (under development - seems not to be working correctly.)

  SBx (patternloop, x>0?count:start)
  SCx (notecut in   x ticks)
  SDx (notedelay    x ticks)
  SEx (patterndelay x lines)

  MBx (NON-STANDARD: reverb)
  MCx (NON-STANDARD: chorus)

  Txx (set tempo)

  By definition, rows played in minute is 24 * tempo / ticks.
  That means, in a second, it is 2/5*tempo/ticks.
  Example: T7D, A06: tick length is 5 / (125+125) = 1 / 50 seconds,
                     six ticks in row, means row is 6 / 50 seconds,
                     25/3 = 8.333.. rows per second.
  Timer ticks per row = rate * speed * 5 / tempo / 2

  Exx and Fxx:

    For Exx, FreqSlide =  xx
    For Fxx, FreqSlide = -xx

    In each frame (there are Axx frames in a row)
      the Period of note will be increased by FreqSlide.
    The output Herz is Period * C2SPD of instrument / 22000.

  Hxy and Kxy:
  
    For Kxy: first do Dxy and then set xy to be the last Hxy xy.

    Vibrato amplitude: y/16       semitones
    Vibrato speed:     x*ticks/64 cycles in second
    Ticks is the value set by Axx -command.

  To make:

    Rxy
    Gxx
    Lxy

  To fix:

    DFx (somehow... grain? karkea)
    DxF (this too)


    EEx
    Exx
    FFx
    FEx
    Fxx
    Hxx
    Kxx
    Qxy
    SEx (does not appear to handle command repeatings correctly)
    Uxx

*/

volatile struct MusData MusData;

static void M_NoteOff(int ch)
{
    MusData.driver->noteoff(ch);
}

static void M_MidiCtrl(int ch, int ctrl, int value)
{
    MusData.driver->midictrl(ch, ctrl, value);
}

static void M_Pan(int ch, int value)
{
    /* Pan limits: -128..+127 */
    MusData.driver->setpan(ch, value - 128);
}

static void M_HzTouch(int ch, int hz)
{
    MusData.driver->hztouch(ch, hz);
}

static void M_GmBend(int ch, int val)
{
    MusData.driver->bend(ch, val);
}

static int UnShiftVol(long v)
{
	return v / (1 << volshift);
}

static long ShiftVol(int v)
{
	return v * (1 << volshift);
}

static void M_NoteOn(int ch, int note, long vol, unsigned Instru, int Herz)
{
    /* M_GmBend(ch, 0x2000);  * redundant */
    MusData.driver->patch(
      ch,
      MusData.Instr[Instru]->D,
      MusData.Instr[Instru]->GM,
      MusData.Instr[Instru]->Bank & 127);
    
    MusData.driver->noteon(
      ch,
      note,
      Herz,
      UnShiftVol(vol) * (MusData.Instr[Instru]->D[8] >> 2) / 63,
      UnShiftVol(vol));
}

static void M_Touch(int ch, unsigned Instru, long vol)
{
    MusData.driver->touch(
      ch,
      UnShiftVol(vol) * (MusData.Instr[Instru]->D[8] >> 2) / 63,
      UnShiftVol(vol));
}

static void PlayLine(void)
{
    struct MusPlayerPos ToSave;
    BYTE ByteRead, Cmd, Puh, Note, Instru, Info, Oct;
    int Restore, Cut, Break;
    int CutLine, BreakOrder, CurChn;
    BYTE GotCmd[MAXCHN];

    if(MusData.PatternDelay)
    {
        for(Cmd=0; Cmd<MAXCHN; MusData.CurPuh[Cmd++] = 255);
        MusData.PatternDelay--;
        return;
    }

    ToSave = MusData.posi;

    for(Cmd=Restore=Cut=Break=0;
        Cmd<MAXCHN; GotCmd[Cmd++]=0);

    if(MusData.PendingSB0)
    {
        /* Do saving at the pattern first line! */
        MusData.saved = ToSave;
        MusData.PendingSB0 = 0;
    }

    if(MusData.posi.Wait)
        --MusData.posi.Wait;

    if(!MusData.posi.Wait)
    {
        /* Iso for-lause, yk */
        for(;;)
        {
            ByteRead = *MusData.posi.Pos++;
            if(ByteRead < 16)break;
    
            Puh    = 255;
            CurChn = ByteRead&15; /* There are no hdr.channels anymore */
            Instru = MusData.CurInst[CurChn];
            Info = 0;
    
            if(ByteRead & 16)
                Info = *MusData.posi.Pos++;
    
            if(ByteRead & 32) /* Note(+instru) specified? */
            {
                BYTE a = MusData.InstNum[CurChn];
                Puh = *MusData.posi.Pos++;
    
                if(!(MusData.Hdr->OrdNum&256)) /* What does this mean? */
                {
    GetInsNum:      MusData.InstNum[CurChn] = a = *MusData.posi.Pos++;
                    /* Normaali tilanne */
                }
                else if(Puh&128)
                {
                    if(Puh<254)Puh&=127;
                    goto GetInsNum;
                }
                if(a)
                {
                    MusData.CurInst[CurChn] = Instru = a;
                    MusData.CurVol[CurChn] = ShiftVol(MusData.Instr[Instru]->Volume & 63);
                }
                Oct = Puh>>4;
                Note= Puh&15;
            }
    
            if(ByteRead & 64) /* Volume specified? */
            {
                BYTE a = *MusData.posi.Pos++;
                MusData.CurVol[CurChn] = ShiftVol(a);
                MusData.NoteDelay[CurChn] = 1;
            }
    
            Cmd = '-';
            if(ByteRead & 128) /* Cmd(+infobyte) specified? */
                GotCmd[CurChn] = Cmd = (*MusData.posi.Pos++) | 64;
    
            if(Puh != 255)
            {
                MusData.NoteCut[CurChn] = 0;
                MusData.NoteDelay[CurChn] = 1;
            }
    
            if((MusData.CurPuh[CurChn]=Puh) < 254)
            {
                int r;
    
                MusData.NNum[CurChn] = (12*Oct) + Note;
                MusData.NNum[CurChn] += MusData.Instr[Instru]->FT - 2;
                /* MIDI finetuning. -2 is a hack. */
    
                r = ((MusData.Instr[Instru]->D[9] >> 2) & 15);
    
                if(r)
                {
                    r = rand() % (r + 1);
                    MusData.NoteDelay[CurChn] = r+1;
                }
    
                if(MusData.Instr[Instru]->Volume&128) /* AdLib finetuning also? */
                {
                    Note = MusData.NNum[CurChn]%12;
                    Oct  = MusData.NNum[CurChn]/12;
                }
                
                MusData.Herz[CurChn] = /* Period is defined in m_opl.c */
                    (unsigned long)(Period[Note] * MusData.Instr[Instru]->C2Spd
                                                 / 22000L) << Oct;
            }
    
            switch(Cmd)
            {
                case 'A':
                    MusData.FrameCount = Info;
                    break;
                case 'T':
                    MusData.Tempo = Info;
                    break;
                case 'J':
                    if(Info)MusData.Arpeggio[CurChn] = Info;
                    break;
                case 'P':
                    if(Info)
                        MusData.ArpegSpd = Info;
                    break;
                case 'Q':
                    if(Info)
                    {
                    	MusData.Retrig[CurChn] = Info;
                    }
                    break;
                case 'B':
                    Break = 1;
                    BreakOrder = Info;
                    break;
                case 'X':
                {
                    M_Pan(CurChn, Info);
                    break;
                }
    
                case 'D':
                case 'K':
                    if((ByteRead & 64) && !MusData.CurVol[CurChn])
                    {
                        /* FIXME: Why is this here? */
                        MusData.CurVol[CurChn] = ShiftVol(1);
                    }
    
                    if(!Info)Info = MusData.DxxDefault[CurChn];
                    MusData.DxxDefault[CurChn] = Info;
    
                    if(Info==0xFF)MusData.CurVol[CurChn] = ShiftVol(63);
                                                             /* DFF, maximize */
                    else if(Info > 0xF0)                     /* DFx           */
                    {
                        MusData.VolSlide[CurChn] = -(int)(Info&0x0F);
                        MusData.FineVolSlide[CurChn] = 1;
                    }
                    else if((Info&0x0F)==0x0F && (Info&0xF0)) /* DxF */
                    {
                        MusData.VolSlide[CurChn] = Info>>4;
                        MusData.FineVolSlide[CurChn] = 1;
                    }
                    else if(!Info);                           /* D00 */
                    else if(!(Info&0xF0))                     /* D0x */
                    {
                        MusData.VolSlide[CurChn] = -(Info&0x0F);
                        MusData.FineVolSlide[CurChn] = 0;
                    }
                    else if(!(Info&0x0F))                     /* Dx0 */
                    {
                        MusData.VolSlide[CurChn] = Info>>4;
                        MusData.FineVolSlide[CurChn] = 0;
                    }
                    if(Cmd=='K'){Info=0; goto CmdH;}
                    break;
                case 'E':
                    MusData.FineFrqSlide[CurChn] = 0;
                    if(!Info)Info=MusData.ExxDefault[CurChn];
                    MusData.ExxDefault[CurChn]=Info;
                    if(Info>=0xF0)MusData.FineFrqSlide[CurChn]=1, Info&=15;
                    MusData.FreqSlide[CurChn] = -Info;
                    break;
                case 'N':
                    /* Bend limits: 0x0000 - 0x3FFF
                     * 0x80 << 6 = 0x2000
                     * 0x00 << 6 = 0x0000
                     * 0xFF << 6 = 0x3FE0 or similar
                     */
                    M_GmBend(CurChn, (unsigned)Info << 6);
                    break;
                case 'F':
                    if(!Info)Info=MusData.FxxDefault[CurChn];
                    MusData.FxxDefault[CurChn]=Info;
                    MusData.FineFrqSlide[CurChn] = 0;
                    if(Info>=0xF0)MusData.FineFrqSlide[CurChn]=1, Info&=15;
                    MusData.FreqSlide[CurChn] = Info;
                    break;
                case 'H':
                {
    CmdH:           if(!Info)Info=MusData.HxxDefault[CurChn];
                    MusData.HxxDefault[CurChn]=Info;
                    if(Info & 0x0F)MusData.VibDepth[CurChn] = Info & 0x0F;
                    if(Info & 0xF0)MusData.VibSpd  [CurChn] =(Info & 0xF0) >> 2;
                    MusData.VibPos[CurChn] += MusData.VibSpd[CurChn];
                    
                    M_MidiCtrl(CurChn, 1,
                        (MusData.VibDepth[CurChn]+
                         MusData.VibSpd  [CurChn]) << 2);

                    break;
                }
                case 'C':
                    Cut = 1;
                    CutLine = (Info>>4)*10 + (Info&15);
                    break;
                case 'M':
                    if(!Info)Info = MusData.MxxDefault[CurChn];
                    MusData.MxxDefault[CurChn] = Info;
                    if((Info>>4)==0xB) /* reverb */
                    {
                        M_MidiCtrl(CurChn, 91, (Info&15)<<3);
                    }
                    else if((Info>>4)==0xC) /* chorus */
                    {
                        M_MidiCtrl(CurChn, 93, (Info&15)<<3);
                    }
                    break;
                case 'S':
                    if(!Info)Info = MusData.SxxDefault[CurChn]; /* Notice */
                    MusData.SxxDefault[CurChn] = Info;          /* Notice */

                    if(Info == 0xB0)
                    {
                        /* ToSave is what we had on the beginning of line */
                        MusData.saved = ToSave;
                    }
                    else if((Info > 0xB0)&&(Info <= 0xBF))
                    {
                        if(MusData.LoopCount == 0)
                            MusData.LoopCount = Info&0x0F;
                        else
                            MusData.LoopCount--;

                        if(MusData.LoopCount)
                        {
                            /* RestorePos() immediately after this line */
                            Restore = 1;
                        }
                        else
                        {
                            /* SavePos() at next line */
                            MusData.PendingSB0 = 1;
                        }
                    }
                    else if((Info>>4)==0xC)
                    {
                        /* Mitkhn ideaa olisi komennossa SC0... */
                        MusData.NoteCut[CurChn] = 1 + (Info & 15);
                    }
                    else if((Info>>4)==0xD)
                    {
                        /* Mitkhn ideaa olisi komennossa SD0... */
                        MusData.NoteDelay[CurChn] = 1 + (Info & 15);
                    }
                    else if((Info>>4)==0xE)
                    {
                        /* Mitkhn ideaa olisi komennossa SE0... */
                        MusData.PatternDelay = Info - 0xE0;
                    }
            }
        }
        if(ByteRead==15)MusData.posi.Wait = *MusData.posi.Pos++;
        else if(ByteRead>0)MusData.posi.Wait=ByteRead+1;
    }

    for(CurChn=0; CurChn<MAXCHN; CurChn++)
    {
        Cmd = GotCmd[CurChn];

        if(Cmd != 'D' && Cmd != 'K')MusData.VolSlide[CurChn] = 0;
        if(Cmd != 'E' && Cmd != 'F')MusData.FreqSlide[CurChn] = 0;
        if(Cmd != 'J')MusData.Arpeggio[CurChn]=0;
        if(Cmd != 'Q')MusData.Retrig[CurChn]  =0;
        if(Cmd != 'H' && Cmd != 'K')
        {
            if(MusData.VibDepth[CurChn]
            || MusData.VibSpd[CurChn])
                M_MidiCtrl(CurChn, 1, 0);

            MusData.VibDepth[CurChn] =
            MusData.VibSpd  [CurChn] =
            MusData.VibPos  [CurChn] = 0;
        }
    }
    if(Restore)
    {
        /* If got SBx _now_ */
        MusData.posi = MusData.saved;
        return;
    }
    if(!MusData.PatternDelay)MusData.posi.Row++;

    if(!Cut && !Break
    && MusData.posi.Pos >= MusData.Patn[MusData.posi.Pat].Ptr
                         + MusData.Patn[MusData.posi.Pat].Len)
    {
        MusData.posi.Ord = (MusData.posi.Ord+1) % (BYTE)MusData.Hdr->OrdNum;
        goto P3;
    }

    if(Break)
    {
        MusData.posi.Ord = BreakOrder;
P3:     if(!Cut)CutLine = 0;
        goto P2;
    }
    if(Cut)
    {
        MusData.posi.Ord = (MusData.posi.Ord+1) % (BYTE)MusData.Hdr->OrdNum;
P2:     MusData.posi.Pat = MusData.Orders[MusData.posi.Ord];
        MusData.posi.Pos = MusData.Patn[MusData.posi.Pat].Ptr;
        MusData.posi.Row = CutLine;
        while(CutLine)
        {
            if(MusData.posi.Wait)MusData.posi.Wait--;
            if(!MusData.posi.Wait)
            {
                for(;;)
                {
                    ByteRead = *MusData.posi.Pos++;
                    if(ByteRead<16)break;
                    if(ByteRead&16)MusData.posi.Pos++;
                    if(ByteRead&32)
                    {
                        BYTE a = *MusData.posi.Pos++;
                        if((!(MusData.Hdr->OrdNum&256)) || (a&128))
                            MusData.posi.Pos++;
                    }
                    if(ByteRead&64)MusData.posi.Pos++;
                    if(ByteRead&128)MusData.posi.Pos++;
                }
                if(ByteRead==15)MusData.posi.Wait = *MusData.posi.Pos++;
                else if(ByteRead>0)MusData.posi.Wait=ByteRead+1;
            }
            CutLine--;
        }
        /* At this point, we are always on a new order. */
        MusData.PendingSB0 = 1;
    }
}

/* Handles one timer tick */
/* Called from m_irq.c    */
void TimerHandler(void)
{
    int ch;
    int frametick = 0;
    
    if(MusData.Paused || !MusData.driver)return;
    
    if(++MusData.FrameTimer >= MusData.FrameLimit)
    {
    	MusData.FrameTimer = 0;
    	if(++MusData.Frame >= MusData.FrameCount)
    	{
	        PlayLine();
	        MusData.Frame      = 0;
	        MusData.FrameLimit = Rate * 5 / MusData.Tempo / 2;
    	}
    	frametick = 1;
    }
    
    for(ch=0; ch<MAXCHN; ch++)
    {
        if(frametick)
        {
            if(MusData.NoteCut[ch] > 0)
            {
            	if(!--MusData.NoteCut[ch])
	                M_NoteOff(ch);
            }

            if(MusData.NoteDelay[ch] > 0)
            {
            	if(!--MusData.NoteDelay[ch])
	                switch(MusData.CurPuh[ch])
    	            {
	                    case 255:   /* Tyhj rivi */
    	                    M_Touch(ch, MusData.CurInst[ch],
    	                                MusData.CurVol[ch]);
        	                break;
            	        case 254:   /* ^^ -rivi */
	                        M_NoteOff(ch);
    	                    break;
        	            default:    /* Rivi jolla nuotti */
            	            M_NoteOff(ch);
	                        M_NoteOn(ch, MusData.NNum[ch],
	                                     MusData.CurVol[ch],
	                                     MusData.CurInst[ch],
	                                     MusData.Herz[ch]);
                	}
            	/* FIXME:
            	   When PatternDelay is used, NoteDelay
            	   should retrigger the same note again
            	   and again.
            	    !MusData.PatternDelay
            	*/
            }
            
            /* Don't retrig on the first frame */            
            if(MusData.Retrig[ch] && MusData.Frame)
            {
            	unsigned space = MusData.Retrig[ch] & 15;
            	unsigned fade  = MusData.Retrig[ch] >> 4;
            	
                if(!(MusData.Frame % space))
                {
                    MusData.CurVol[ch] -= ShiftVol(fade);
                    if(MusData.CurVol[ch] < 0)MusData.CurVol[ch] = 0;
                    
                    M_NoteOff(ch);
                    M_NoteOn(ch, MusData.NNum[ch],
                                 MusData.CurVol[ch],
                                 MusData.CurInst[ch],
                                 MusData.Herz[ch]);
                }
            }
        }
        
        /* Arpeggio is frame independent */
        /* TODO: FIXME */
        if(MusData.Arpeggio[ch])
        {
            WORD ArpCounter = (((unsigned long)(MusData.ArpegSpd & 127)) << 16) / Rate;
            
            MusData.ArpegClk += ArpCounter;
            if(MusData.ArpegClk < ArpCounter) /* was overflow? */
            {
                int notemp, Note;
            	unsigned long Herz;
            	BYTE Instru = MusData.CurInst[ch];
            	
                switch((MusData.ArpegCnt[ch]++) % (3-(MusData.ArpegSpd>>7)))
                {
                    case 0:
                        Note = 0;
                        break;
                    case 1:
                        Note = MusData.Arpeggio[ch] & 15;
                        break;
                    default:
                        Note = MusData.Arpeggio[ch] >> 4;
                }

                Note += MusData.NNum[ch];
                
                notemp = Note;

                if(!(MusData.Instr[Instru]->Volume&128))
                {
                    notemp -= MusData.Instr[Instru]->FT;
                    /* Remove finetuning effect for AdLib */
                }
                Herz = (unsigned long)(Period[notemp % 12]
                      * MusData.Instr[Instru]->C2Spd / 22000L)
                      << (Note/12);

                M_NoteOff(ch);
                M_NoteOn(ch, Note,
                  MusData.CurVol[ch],
                  Instru, Herz);
        	}
        }
        
        if(MusData.VolSlide[ch])
        {
            long Vol = MusData.CurVol[ch];
            
            if(MusData.FineVolSlide[ch])
                Vol += ShiftVol(MusData.VolSlide[ch]);
                
            if(Vol < 0)Vol = 0;
            if(Vol > ShiftVol(63))Vol = ShiftVol(63);

            MusData.CurVol[ch] = Vol;
            
            M_Touch(ch,
              MusData.CurInst[ch],
              MusData.CurVol[ch]);

            if(MusData.FineVolSlide[ch] && !MusData.PatternDelay)
                MusData.VolSlide[ch] = 0;
        }

        #if 0
        if(MusData.FreqSlide[ch])
        {
            M_HzTouch(ch,
                MusData.Herz[ch] +=
                (((long)(MusData.FreqSlide[ch]
                     * ((MusData.FrameCount)-1))<<14) / Limit) >> 12
            );

            /* M_GmBend(ch, 0x2000); FIXME - not implemented. */

            if(MusData.FineFrqSlide[ch] && !MusData.PatternDelay)
                MusData.FreqSlide[ch]=0;
        }

        /* Vibrato - not correctly working yet */
        if(MusData.Herz[ch])
        {
            if(MusData.VibDepth[ch])
            {
                static const BYTE avibtab[128] =
                {   0,1,3,4,6,7,9,10,12,14,15,17,18,20,21,23,
                    24,25,27,28,30,31,32,34,35,36,38,39,40,41,42,44,
                    45,46,47,48,49,50,51,52,53,54,54,55,56,57,57,58,
                    59,59,60,60,61,61,62,62,62,63,63,63,63,63,63,63,
                    64,63,63,63,63,63,63,63,62,62,62,61,61,60,60,59,
                    59,58,57,57,56,55,54,54,53,52,51,50,49,48,47,46,
                    45,44,42,41,40,39,38,36,35,34,32,31,30,28,27,25,
                    24,23,21,20,18,17,15,14,12,10,9,7,6,4,3,1
                };

                int VibDpt, VibVal = avibtab[MusData.VibPos[ch] & 127];
    
                if(MusData.VibPos[ch] & 0x80)VibVal = -VibVal;
                VibDpt = MusData.VibDepth[ch] << 8;
                VibVal = (VibVal*VibDpt) >> 12;
                
                M_HzTouch(ch, MusData.Herz[ch] - VibVal*8);
    
                MusData.VibPos[ch]++;
            }
        }
        #endif
    }
}

void StartS3M(const BYTE *s)
{
    BYTE a;

    StopS3M();

    /* Pointers are nice :) */
    MusData.Hdr = (InternalHdr *)__FILE__" v"VERSION"(C) 1992,2000 Bisqwit (bisqwit@iki.fi)";
    /* Please don't modify the text above. It would be nice to   *
     * see my name inside games made by people unknown to me :-) */

    MusData.Playing = s;
    if(s[1]>2)for(;*s++;); /* Skip possible song name */

    MusData.Hdr = (const InternalHdr *)s;
    s += sizeof(InternalHdr);
    if(MusData.Hdr->PatNum & 512)
        s++; /* ChanCount present */

    MusData.Orders = (const BYTE *)s;
    s += (BYTE)MusData.Hdr->OrdNum;
    
    for(a=0; a<MusData.Hdr->InsNum; a++)
    {
        InternalSample *tmp = (InternalSample *)s;
        
        if(MusData.Hdr->PatNum & 512)
            s += 1 + strlen(tmp->insname);
       
        s += sizeof(*tmp);
        
        MusData.Instr[a+1] = tmp;
    }

    /* This disables some bugs. */
    MusData.Instr[0] = MusData.Instr[1];

    ((InternalHdr *)MusData.Hdr)->PatNum &= ~512;

    for(a=0; a<MusData.Hdr->PatNum; a++)
    {
        MusData.Patn[a].Len = ((BYTE)s[0]) | (((BYTE)s[1]) << 8);
        MusData.Patn[a].Ptr = (s += 2);
        s += MusData.Patn[a].Len;
    }

    for(a=0; a<MAXCHN; a++)
    {
        MusData.CurVol[a] = MusData.CurInst[a] = 0;
        MusData.NoteDelay[a] = -1;
    }
    
    SetPause(1);

    MusSeekOrd(0);
    MusData.LinearMidiVol = (MusData.Hdr->InSpeed & 128) == 128;
    MusData.Tempo      = MusData.Hdr->InTempo;
    MusData.FrameCount = MusData.Hdr->InSpeed & 127;
    MusData.ArpegSpd     = 50;
    MusData.GlobalVolume = 64;
    
    SetPause(0);
}

/* From m_arch.c */
extern int FromArchive;
extern char *Archived;

void StopS3M(void)
{
    if(Archived && !FromArchive)
    {
        free(Archived);
        Archived = NULL;
    }
    
    SetPause(1);
    
    ReleaseINT();
}

void SetPause(int state)
{
    if(state == MusData.Paused)return;
    MusData.Paused = state;
    
    if(MusData.driver)
        MusData.driver->reset();

    if(state) /* pause */
    {
        /* actually nothing to do */
    }
    else      /* unpause */
    {
        InstallINT();
    }
}

void PauseS3M(void)
{
    if(MusData.Paused)
        SetPause(0);
    else
        SetPause(1);
}

void MusSeekOrd(int Num)
{    
    int p = MusData.Paused;
    
    SetPause(1);
    
    /* The player must not modify these at same time */
    MusData.PatternDelay = 0;
    MusData.posi.Wait = 0;
    MusData.posi.Row = 0;
    MusData.posi.Ord = Num;
    MusData.posi.Pat = MusData.Orders[Num];
    MusData.posi.Pos = MusData.Patn[MusData.posi.Pat].Ptr;
    MusData.LoopCount = 0;
    MusData.PendingSB0 = 1;
    MusData.FrameLimit = 1;
    MusData.FrameTimer = 0;
    MusData.Frame      = MusData.FrameCount-1;
    
    SetPause(p);
}

void MusSeekNextOrd(void)
{
    if(MusData.posi.Ord < MusData.Hdr->OrdNum - 1)
        MusSeekOrd(MusData.posi.Ord+1);
}

void MusSeekPrevOrd(void)
{
    if(MusData.posi.Ord > 0)MusSeekOrd(MusData.posi.Ord - 1);
}
